import os
import pandas as pd
import streamlit as st
from acv_explainers.acv_tree import ACVTreeAgnostic, ACVTree
from acv_app import sdp_app
from acv_app import sv_app
from joblib import dump, load
import argparse
import subprocess

DATA_PATH = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))), 'data')
# LOAD_PATH = '/home/samoukou/Documents/INVASE-master'
APP_PATH = __file__


# Set page title
# st.set_page_config(page_title='ACPR challenge', layout="wide")

# define all required cache data for the app

def compile_acv(model, x_train, y_train, x_test, y_test, path):
    if type(x_train) != pd.core.frame.DataFrame:
        x_train = pd.DataFrame(x_train, columns=['X{}'.format(i) for i in range(x_train.shape[1])])
        x_test = pd.DataFrame(x_test, columns=['X{}'.format(i) for i in range(x_test.shape[1])])

    if type(y_train) != pd.core.frame.DataFrame:
        y_test = pd.DataFrame(y_test)
        y_train = pd.DataFrame(y_train)

    acvtree = ACVTree(model, x_train.values)
    acvtree_agnostic = ACVTreeAgnostic(model, x_train.values[:5])

    dump(acvtree, os.path.join(path, 'acvtree.joblib'))
    dump(acvtree_agnostic, os.path.join(path, 'acvtree_agnostic.joblib'))
    x_train.to_csv(os.path.join(path, 'x_train.csv'), index=False)
    y_train.to_csv(os.path.join(path, 'y_train.csv'), index=False)
    x_test.to_csv(os.path.join(path, 'x_test.csv'), index=False)
    y_test.to_csv(os.path.join(path, 'y_test.csv'), index=False)


def compile_ACXplainers(ACXplainers, x_train, y_train, x_test, y_test, path):
    acvtree = ACXplainers
    acvtree_agnostic = ACXplainers

    if type(x_train) != pd.core.frame.DataFrame:
        x_train = pd.DataFrame(x_train, columns=['X{}'.format(i) for i in range(x_train.shape[1])])
        x_test = pd.DataFrame(x_test, columns=['X{}'.format(i) for i in range(x_test.shape[1])])

    if type(y_train) != pd.core.frame.DataFrame:
        y_test = pd.DataFrame(y_test)
        y_train = pd.DataFrame(y_train)

    dump(acvtree, os.path.join(path, 'acvtree.joblib'))
    dump(acvtree_agnostic, os.path.join(path, 'acvtree_agnostic.joblib'))
    x_train.to_csv(os.path.join(path, 'x_train.csv'), index=False)
    y_train.to_csv(os.path.join(path, 'y_train.csv'), index=False)
    x_test.to_csv(os.path.join(path, 'x_test.csv'), index=False)
    y_test.to_csv(os.path.join(path, 'y_test.csv'), index=False)


def run_webapp(pickle_path):
    run_file = subprocess.run(["streamlit", "run", APP_PATH, "--", "--path", pickle_path])
    return run_file


def main(LOAD_PATH):
    """Loads data, preprocess it, creates a simple model then wrapps the data and the model in a Xplainer
    Once data and the model are wrapped, it creates the streamlit front PAGES
    """

    st.set_page_config(page_title='ACV demo', layout="wide")

    PAGES = {
        "SDP based explanations": sdp_app,
        "Shapley based Explanations": sv_app
    }

    st.sidebar.title("Choose your explanations")
    persona = st.sidebar.radio("", list(PAGES.keys()))

    @st.cache
    def load_data():
        x_train = pd.read_csv(os.path.join(LOAD_PATH, 'x_train.csv'))
        y_train = pd.read_csv(os.path.join(LOAD_PATH, 'y_train.csv')).values.squeeze()
        x_test = pd.read_csv(os.path.join(LOAD_PATH, 'x_test.csv'))
        y_test = pd.read_csv(os.path.join(LOAD_PATH, 'y_test.csv')).values.squeeze()
        return x_train, x_test, y_train, y_test

    # data_load_state = st.sidebar.text('Loading data...')
    x_train, x_test, y_train, y_test = load_data()

    # data_load_state.text("Load data Done !")

    @st.cache(allow_output_mutation=True)
    def load_model():
        acvtree = load(os.path.join(LOAD_PATH, 'acvtree_agnostic.joblib'))
        return acvtree

    # model_load_state = st.sidebar.text('Loading model...')
    acvtree = load_model()
    # model_load_state.text("Load model Done!")

    if persona == 'Shapley based Explanations':
        st.title('Local Explanations based on Shapley values ')
        acvtree = load(os.path.join(LOAD_PATH, 'acvtree.joblib'))
    else:
        st.title('Consistent Local Explanations based on SDP ')

    page = PAGES[persona]
    page.write_pg(x_train, x_test, y_train, y_test, acvtree)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Params of ACV Explanations Web App")
    parser.add_argument(
        '--path',
        default='./',
        type=str)

    args = parser.parse_args()

    main(args.path)



# cython: linetrace=True


from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy

from Cython.Compiler.Options import get_directive_defaults
directive_defaults = get_directive_defaults()

directive_defaults['linetrace'] = True
directive_defaults['binding'] = True

setup(cmdclass={'build_ext': build_ext},
      ext_modules=[Extension('cyext_acv', ['cyext_acv.pyx'])],
      include_dirs=[numpy.get_include()],
      define_macros=[('CYTHON_TRACE_NOGIL', '1')]
)


# distutils: language = c++

from libcpp.vector cimport vector
import numpy as np
cimport numpy as np
ctypedef np.float64_t double
cimport cython
from scipy.special import comb
import itertools
from tqdm import tqdm
from cython.parallel cimport prange, parallel, threadid
cimport openmp

cdef extern from "<algorithm>" namespace "std" nogil:
     iter std_remove "std::remove" [iter, T](iter first, iter last, const T& val)
     iter std_find "std::find" [iter, T](iter first, iter last, const T& val)

cdef extern from "limits.h":
    unsigned long ULONG_MAX


cpdef double binomialC(unsigned long N, unsigned long k) nogil:
    cdef double r
    r = _comb_int_long(N, k)
    if r != 0:
        return r

cpdef double _comb_int_long(unsigned long N, unsigned long k) nogil:
    """
    Compute binom(N, k) for integers.
    Returns 0 if error/overflow encountered.
    """
    cdef double val
    cdef unsigned long long j, M, nterms

    if k > N or N == ULONG_MAX:
        return 0

    M = N + 1
    nterms = min(k, N - k)

    val = 1

    for j in range(1, nterms + 1):
        # Overflow check
        # if val > ULONG_MAX // (M - j):
        #   return 0

        val *= M - j
        val //= j

    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef leaves_cache(
    const double[:, :] data,
    const double[:, :, :] values,
    const double[:, :, :, :] partition_leaves_trees,
    const long[:, :] leaf_idx_trees,
    const long[::1] leaves_nb,
    const long scaling,
    const vector[vector[vector[long]]] node_idx_trees,
    const vector[vector[long]] C, int num_threads):


    cdef unsigned int d = values.shape[2]
    cdef unsigned int m = data.shape[1]
    cdef unsigned int n_trees = values.shape[0]
    cdef unsigned int max_leaves = partition_leaves_trees.shape[1]

    cdef long[:, :, :, ::1] S
    S = np.zeros((m, 2**scaling, max_leaves, m), dtype=np.int)

    cdef unsigned int a_it, nb_leaf, va,  counter, ci, cj, ite, pow_set_size, nv, ns, add
    cdef unsigned int b, leaf_numb, i, s, j, i1, i2, l, na_bool, o_all, csi, cs
    cdef double[:, :] lm
    lm = np.zeros((n_trees, max_leaves))

    cdef vector[vector[long]] C_buff, va_id, buff
    cdef vector[long] coal_va, remove_va, node_id_b, buff_l
    buff.resize(max_leaves)
    for i in range(max_leaves):
        buff[i].resize(1)

    cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
    node_id_v2.resize(max_leaves)
    C_b.resize(max_leaves)
    Sm.resize(max_leaves)
    node_id.resize(max_leaves)

    cdef vector[vector[vector[vector[long]]]] node_id_va
    node_id_va.resize(m)
    for i in range(m):
        node_id_va[i].resize(max_leaves)

    cdef int S_size, va_size,  b_it, nv_bool

    cdef long set_size

    if C[0].size() != 0:
        C_buff = C
        for ci in range(C.size()):
            for cj in range(C[ci].size()):
                coal_va.push_back(C[ci][cj])

        for i in range(m):
            if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
                remove_va.push_back(i)
                buff[0][0] = i
                va_id.push_back(buff[0])
            else:
                for ci in range(C_buff.size()):
                    if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
                        va_id.push_back(C_buff[ci])
                        std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
                        C_buff.pop_back()
                        break

    else:
        for i in range(m):
            buff[0][0] = i
            va_id.push_back(buff[0])

    cdef double p_s, p_si, coef, coef_0
    cdef double[:, :, :, :] lm_s, lm_si
    lm_s = np.zeros((n_trees, va_id.size(), max_leaves, 2**scaling))
    lm_si = np.zeros((n_trees, va_id.size(), max_leaves, 2**scaling))

    for b in range(n_trees):
        nb_leaf = leaves_nb[b]
        for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
            node_id_v2[leaf_numb].clear()
            lm[b, leaf_numb] = 0
            for i in range(data.shape[0]):
                a_it = 0
                for s in range(data.shape[1]):
                    if (data[i, s] <= partition_leaves_trees[b, leaf_numb, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_numb, s, 0]):
                        a_it = a_it + 1
                if a_it == data.shape[1]:
                    lm[b, leaf_numb] = lm[b, leaf_numb] + 1

            if C[0].size() != 0:
                C_b[leaf_numb] = C
                for nv in range(node_idx_trees[b][leaf_numb].size()):
                    add = 0
                    for ns in range(remove_va.size()):
                        if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
                            add = 1
                            buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
                            node_id_v2[leaf_numb].push_back(buff[leaf_numb])
                            break
                    if add == 0:
                        for ci in range(C_b[leaf_numb].size()):
                            for cj in range(C_b[leaf_numb][ci].size()):
                                if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
                                    add = 1
                                    node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
                                    break
                            if add == 1:
                                std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
                                C_b[leaf_numb].pop_back()
                                break
                node_id[leaf_numb] = node_id_v2[leaf_numb]
            else:
                node_id[leaf_numb].clear()
                for i in range(node_idx_trees[b][leaf_numb].size()):
                    buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
                    node_id[leaf_numb].push_back(buff[leaf_numb])


            for va in range(va_id.size()):
                node_id_va[va][leaf_numb] = node_id[leaf_numb]
                if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
                    continue

                std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
                node_id_va[va][leaf_numb].pop_back()

                set_size = node_id_va[va][leaf_numb].size()
                pow_set_size = 2**set_size

                for counter in range(0, pow_set_size):
                    va_size = 0
                    S_size = 0
                    for ci in range(set_size):
                        if((counter & (1 << ci)) > 0):
                            for cj in range(node_id_va[va][leaf_numb][ci].size()):
                                S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
                                S_size = S_size + 1
                            va_size = va_size + 1

                    for i in range(data.shape[0]):
                        b_it = 0
                        for s in range(S_size):
                            if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
                                b_it = b_it + 1

                        if b_it == S_size:
                            lm_s[b, va, leaf_numb, counter] = lm_s[b, va, leaf_numb, counter] + 1

                            nv_bool = 0
                            for nv in range(va_id[va].size()):
                                if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
                                    nv_bool = nv_bool + 1
                                    continue

                            if nv_bool == 0:
                                lm_si[b, va, leaf_numb, counter] = lm_si[b, va, leaf_numb, counter] + 1

    return np.asarray(lm), np.asarray(lm_s), np.asarray(lm_si)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef shap_values_leaves_cache(const double[:, :] X,
    const double[:, :] data,
    const double[:, :, :] values,
    const double[:, :, :, :] partition_leaves_trees,
    const long[:, :] leaf_idx_trees,
    const long[::1] leaves_nb,
    const double [:, :] lm,
    const double[:, :, :, :] lm_s,
    const double[:, :, :, :] lm_si,
    const long scaling,
    const vector[vector[vector[long]]] node_idx_trees,
    const vector[vector[long]] C, int num_threads):


    cdef unsigned int d = values.shape[2]
    cdef unsigned int N = X.shape[0]
    cdef unsigned int m = X.shape[1]
    cdef unsigned int n_trees = values.shape[0]
    cdef unsigned int max_leaves = partition_leaves_trees.shape[1]

    cdef double[:, :, :] phi
    phi = np.zeros((N, m, d))

    cdef double[ :, :, :,  :, :] phi_b
    phi_b = np.zeros((max_leaves, 2**scaling, N, m, d))

    cdef long[:, :, :, ::1] S
    S = np.zeros((m, 2**scaling, max_leaves, m), dtype=np.int)

    cdef unsigned int a_it, nb_leaf, va,  counter, ci, cj, ite, pow_set_size, nv, ns, add
    cdef unsigned int b, leaf_numb, i, s, j, i1, i2, l, na_bool, o_all, csi, cs

    cdef vector[vector[long]] C_buff, va_id, buff
    cdef vector[long] coal_va, remove_va, node_id_b, buff_l
    buff.resize(max_leaves)
    for i in range(max_leaves):
        buff[i].resize(1)

    cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
    node_id_v2.resize(max_leaves)
    C_b.resize(max_leaves)
    Sm.resize(max_leaves)
    node_id.resize(max_leaves)

    cdef vector[vector[vector[vector[long]]]] node_id_va
    node_id_va.resize(m)
    for i in range(m):
        node_id_va[i].resize(max_leaves)

    cdef int S_size, va_size,  b_it, nv_bool

    cdef long set_size

    if C[0].size() != 0:
        C_buff = C
        for ci in range(C.size()):
            for cj in range(C[ci].size()):
                coal_va.push_back(C[ci][cj])

        for i in range(m):
            if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
                remove_va.push_back(i)
                buff[0][0] = i
                va_id.push_back(buff[0])
            else:
                for ci in range(C_buff.size()):
                    if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
                        va_id.push_back(C_buff[ci])
                        std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
                        C_buff.pop_back()
                        break

    else:
        for i in range(m):
            buff[0][0] = i
            va_id.push_back(buff[0])

    cdef double p_s, p_si, coef, coef_0

    for b in range(n_trees):
        for leaf_numb in range(phi_b.shape[0]):
            for counter in range(phi_b.shape[1]):
                for i in range(N):
                    for j in range(m):
                        for i2 in range(d):
                            phi_b[leaf_numb, counter, i, j, i2] = 0
        nb_leaf = leaves_nb[b]
        for leaf_numb in prange(nb_leaf, nogil=True):
            node_id_v2[leaf_numb].clear()

            if C[0].size() != 0:
                C_b[leaf_numb] = C
                for nv in range(node_idx_trees[b][leaf_numb].size()):
                    add = 0
                    for ns in range(remove_va.size()):
                        if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
                            add = 1
                            buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
                            node_id_v2[leaf_numb].push_back(buff[leaf_numb])
                            break
                    if add == 0:
                        for ci in range(C_b[leaf_numb].size()):
                            for cj in range(C_b[leaf_numb][ci].size()):
                                if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
                                    add = 1
                                    node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
                                    break
                            if add == 1:
                                std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
                                C_b[leaf_numb].pop_back()
                                break
                node_id[leaf_numb] = node_id_v2[leaf_numb]
            else:
                node_id[leaf_numb].clear()
                for i in range(node_idx_trees[b][leaf_numb].size()):
                    buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
                    node_id[leaf_numb].push_back(buff[leaf_numb])


            for va in range(va_id.size()):
                node_id_va[va][leaf_numb] = node_id[leaf_numb]
                if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
                    continue

                std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
                node_id_va[va][leaf_numb].pop_back()

                set_size = node_id_va[va][leaf_numb].size()
                pow_set_size = 2**set_size

                for counter in range(0, pow_set_size):
                    va_size = 0
                    S_size = 0
                    for ci in range(set_size):
                        if((counter & (1 << ci)) > 0):
                            for cj in range(node_id_va[va][leaf_numb][ci].size()):
                                S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
                                S_size = S_size + 1
                            va_size = va_size + 1

                    for i in range(N):

                        csi = 0
                        cs = 0

                        o_all = 0
                        for s in range(S_size):
                            if ((X[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_numb, S[va, counter, leaf_numb, s], 0])):
                                o_all = o_all + 1

                        if o_all == S_size:
                            cs = 1
                            nv_bool = 0
                            for nv in range(va_id[va].size()):
                                if ((X[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_numb, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_numb, va_id[va][nv], 0])):
                                    nv_bool = nv_bool + 1
                                    continue

                            if nv_bool == 0:
                                csi = 1
                        coef = 0
                        for l in range(1, m - set_size):
                            coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0

                        coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0

                        if S_size == 0:
                            p_s = lm[b, leaf_numb]/data.shape[0]
                        else:
                            p_s = (cs * lm[b, leaf_numb])/lm_s[b, va, leaf_numb, counter] if lm_s[b, va, leaf_numb, counter] !=0 else 0
                        p_si = (csi * lm[b, leaf_numb])/lm_si[b, va, leaf_numb, counter] if lm_si[b, va, leaf_numb, counter] !=0 else 0

                        for nv in range(va_id[va].size()):
                            for i2 in range(d):
                                phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]

        for i in range(N):
            for j in range(m):
                for i2 in range(d):
                    for leaf_numb in range(phi_b.shape[0]):
                        for counter in range(phi_b.shape[1]):
                            phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]

    return np.asarray(phi)/m

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef shap_values_leaves_normalized_cache(const double[:, :] X,
    const double[:, :] data,
    const double[:, :, :] values,
    const double[:, :, :, :] partition_leaves_trees,
    const long[:, :] leaf_idx_trees,
    const long[::1] leaves_nb,
    const double[:, :, :, :, :] lm,
    const double[:, :, :, :, :] lm_s,
    const double[:, :, :, :, :] lm_si,
    const long scaling,
    const vector[vector[vector[long]]] node_idx_trees,
    const vector[vector[long]] C, int num_threads):


    cdef unsigned int d = values.shape[2]
    cdef unsigned int N = X.shape[0]
    cdef unsigned int m = X.shape[1]
    cdef unsigned int n_trees = values.shape[0]
    cdef unsigned int max_leaves = partition_leaves_trees.shape[1]

    cdef double[:, :, :] phi
    phi = np.zeros((N, m, d))

    cdef double[ :, :, :,  :, :] phi_b
    phi_b = np.zeros((max_leaves, 2**scaling, N, m, d))

    cdef long[:, :, :, ::1] S
    S = np.zeros((m, 2**scaling, max_leaves, m), dtype=np.int)

    cdef unsigned int a_it, nb_leaf, va,  counter, ci, cj, ite, pow_set_size, nv, ns, add
    cdef unsigned int b, leaf_numb, i, s, j, i1, i2, l, na_bool, o_all, leaf_n

    cdef vector[vector[long]] C_buff, va_id, buff
    cdef vector[long] coal_va, remove_va, node_id_b, buff_l
    buff.resize(max_leaves)
    for i in range(max_leaves):
        buff[i].resize(1)

    cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
    node_id_v2.resize(max_leaves)
    C_b.resize(max_leaves)
    Sm.resize(max_leaves)
    node_id.resize(max_leaves)

    cdef vector[vector[vector[vector[long]]]] node_id_va
    node_id_va.resize(m)
    for i in range(m):
        node_id_va[i].resize(max_leaves)

    cdef int S_size, va_size,  b_it, nv_bool

    cdef long set_size

    if C[0].size() != 0:
        C_buff = C
        for ci in range(C.size()):
            for cj in range(C[ci].size()):
                coal_va.push_back(C[ci][cj])

        for i in range(m):
            if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
                remove_va.push_back(i)
                buff[0][0] = i
                va_id.push_back(buff[0])
            else:
                for ci in range(C_buff.size()):
                    if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
                        va_id.push_back(C_buff[ci])
                        std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
                        C_buff.pop_back()
                        break

    else:
        for i in range(m):
            buff[0][0] = i
            va_id.push_back(buff[0])

    cdef double p_s, p_si, coef, coef_0, n_s, n_si, csi, cs, csi_n, cs_n

    for b in range(n_trees):
        for leaf_numb in range(phi_b.shape[0]):
            for counter in range(phi_b.shape[1]):
                for i in range(N):
                    for j in range(m):
                        for i2 in range(d):
                            phi_b[leaf_numb, counter, i, j, i2] = 0
        nb_leaf = leaves_nb[b]
        for leaf_numb in prange(nb_leaf, nogil=True):
            node_id_v2[leaf_numb].clear()

            if C[0].size() != 0:
                C_b[leaf_numb] = C
                for nv in range(node_idx_trees[b][leaf_numb].size()):
                    add = 0
                    for ns in range(remove_va.size()):
                        if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
                            add = 1
                            buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
                            node_id_v2[leaf_numb].push_back(buff[leaf_numb])
                            break
                    if add == 0:
                        for ci in range(C_b[leaf_numb].size()):
                            for cj in range(C_b[leaf_numb][ci].size()):
                                if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
                                    add = 1
                                    node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
                                    break
                            if add == 1:
                                std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
                                C_b[leaf_numb].pop_back()
                                break
                node_id[leaf_numb] = node_id_v2[leaf_numb]
            else:
                node_id[leaf_numb].clear()
                for i in range(node_idx_trees[b][leaf_numb].size()):
                    buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
                    node_id[leaf_numb].push_back(buff[leaf_numb])


            for va in range(va_id.size()):
                node_id_va[va][leaf_numb] = node_id[leaf_numb]
                if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
                    continue

                std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
                node_id_va[va][leaf_numb].pop_back()

                set_size = node_id_va[va][leaf_numb].size()
                pow_set_size = 2**set_size

                for counter in range(0, pow_set_size):
                    va_size = 0
                    S_size = 0
                    for ci in range(set_size):
                        if((counter & (1 << ci)) > 0):
                            for cj in range(node_id_va[va][leaf_numb][ci].size()):
                                S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
                                S_size = S_size + 1
                            va_size = va_size + 1

                    for i in range(N):

                        coef = 0
                        for l in range(1, m - set_size):
                            coef = coef + (1.*binomialC(m - set_size - 1, l))/binomialC(m - 1, l + va_size) if binomialC(m - 1, l + va_size) !=0 else 0

                        coef_0 = 1./binomialC(m-1, va_size) if binomialC(m-1, va_size) !=0 else 0

                        n_s = 0
                        n_si = 0
                        csi = 0
                        cs = 0

                        if S_size == 0:
                            p_s = lm[b, va, leaf_numb, leaf_numb, counter]/data.shape[0]
                            for leaf_n in range(nb_leaf):
                                csi_n = 0
                                cs_n = 0
                                o_all = 0

                                for s in range(S_size):
                                    if ((X[i, S[va, counter, leaf_numb, s]] < partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] >= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
                                        o_all = o_all + 1

                                if o_all == S_size:
                                    cs_n = 1
                                    nv_bool = 0
                                    for nv in range(va_id[va].size()):
                                        if ((X[i, va_id[va][nv]] >= partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] < partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
                                            nv_bool = nv_bool + 1
                                            continue

                                    if nv_bool == 0:
                                        csi_n = 1

                                n_si = n_si + (csi_n * lm[b, va, leaf_numb, leaf_n, counter])/lm_si[b, va, leaf_numb, leaf_n, counter] if lm_si[b, va, leaf_numb, leaf_n, counter] !=0 else 0

                                if leaf_n == leaf_numb:
                                    csi = csi_n
                                    cs = cs_n

                            p_si = (csi * lm[b, va, leaf_numb, leaf_numb, counter])/(lm_si[b, va, leaf_numb, leaf_numb, counter] * n_si) if n_si*lm_si[b, va, leaf_numb, leaf_numb, counter] !=0 else 0
                        else:

                            for leaf_n in range(nb_leaf):
                                csi_n = 0
                                cs_n = 0
                                o_all = 0

                                for s in range(S_size):
                                    if ((X[i, S[va, counter, leaf_numb, s]] < partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (X[i, S[va, counter, leaf_numb, s]] >= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
                                        o_all = o_all + 1

                                if o_all == S_size:
                                    cs_n = 1
                                    nv_bool = 0
                                    for nv in range(va_id[va].size()):
                                        if ((X[i, va_id[va][nv]] >= partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (X[i, va_id[va][nv]] < partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
                                            nv_bool = nv_bool + 1
                                            continue

                                    if nv_bool == 0:
                                        csi_n = 1

                                n_s = n_s + (cs_n * lm[b, va, leaf_numb, leaf_n, counter])/lm_s[b, va, leaf_numb, leaf_n, counter] if lm_s[b, va, leaf_numb, leaf_n, counter] !=0 else 0
                                n_si = n_si + (csi_n * lm[b, va, leaf_numb, leaf_n, counter])/lm_si[b, va, leaf_numb, leaf_n, counter] if lm_si[b, va, leaf_numb, leaf_n, counter] !=0 else 0
                                if leaf_n == leaf_numb:
                                    csi = csi_n
                                    cs = cs_n

                            p_s = (cs * lm[b, va, leaf_numb, leaf_numb, counter])/(lm_s[b, va, leaf_numb, leaf_numb, counter] * n_s) if n_s*lm_s[b, va, leaf_numb, leaf_numb, counter] !=0 else 0
                            p_si = (csi * lm[b, va, leaf_numb, leaf_numb, counter])/(lm_si[b, va, leaf_numb, leaf_numb, counter] * n_si) if n_si*lm_si[b, va, leaf_numb, leaf_numb, counter] !=0 else 0

                        for nv in range(va_id[va].size()):
                            for i2 in range(d):
                                phi_b[leaf_numb, counter, i, va_id[va][nv], i2] += (coef_0 + coef) * (p_si - p_s) * values[b, leaf_idx_trees[b, leaf_numb], i2]

        for i in range(N):
            for j in range(m):
                for i2 in range(d):
                    for leaf_numb in range(phi_b.shape[0]):
                        for counter in range(phi_b.shape[1]):
                            phi[i, j, i2] += phi_b[leaf_numb, counter, i, j, i2]

    return np.asarray(phi)/m

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef leaves_cache_normalized(
    const double[:, :] data,
    const double[:, :, :] values,
    const double[:, :, :, :] partition_leaves_trees,
    const long[:, :] leaf_idx_trees,
    const long[::1] leaves_nb,
    const long scaling,
    const vector[vector[vector[long]]] node_idx_trees,
    const vector[vector[long]] C, int num_threads):


    cdef unsigned int d = values.shape[2]
    cdef unsigned int m = data.shape[1]
    cdef unsigned int n_trees = values.shape[0]
    cdef unsigned int max_leaves = partition_leaves_trees.shape[1]

    cdef long[:, :, :, ::1] S
    S = np.zeros((m, 2**scaling, max_leaves, m), dtype=np.int)

    cdef unsigned int a_it, nb_leaf, va,  counter, ci, cj, ite, pow_set_size, nv, ns, add
    cdef unsigned int b, leaf_numb, i, s, j, i1, i2, l, na_bool, o_all, leaf_n

    cdef vector[vector[long]] C_buff, va_id, buff
    cdef vector[long] coal_va, remove_va, node_id_b, buff_l
    buff.resize(max_leaves)
    for i in range(max_leaves):
        buff[i].resize(1)

    cdef vector[vector[vector[long]]] node_id_v2, C_b, Sm, node_id
    node_id_v2.resize(max_leaves)
    C_b.resize(max_leaves)
    Sm.resize(max_leaves)
    node_id.resize(max_leaves)

    cdef vector[vector[vector[vector[long]]]] node_id_va
    node_id_va.resize(m)
    for i in range(m):
        node_id_va[i].resize(max_leaves)

    cdef int S_size, va_size,  b_it, nv_bool

    cdef long set_size

    if C[0].size() != 0:
        C_buff = C
        for ci in range(C.size()):
            for cj in range(C[ci].size()):
                coal_va.push_back(C[ci][cj])

        for i in range(m):
            if not std_find[vector[long].iterator, long](coal_va.begin(), coal_va.end(), i) != coal_va.end():
                remove_va.push_back(i)
                buff[0][0] = i
                va_id.push_back(buff[0])
            else:
                for ci in range(C_buff.size()):
                    if (std_find[vector[long].iterator, long](C_buff[ci].begin(), C_buff[ci].end(), i) != C_buff[ci].end()):
                        va_id.push_back(C_buff[ci])
                        std_remove[vector[vector[long]].iterator, vector[long]](C_buff.begin(), C_buff.end(), C_buff[ci])
                        C_buff.pop_back()
                        break

    else:
        for i in range(m):
            buff[0][0] = i
            va_id.push_back(buff[0])

    cdef double p_s, p_si, coef, coef_0, n_s, n_si
    cdef double[:, :, :, :, :] lm_s, lm_si, lm_n, csi, cs
    lm_n = np.zeros((n_trees, va_id.size(), max_leaves, max_leaves, 2**scaling))
    lm_s = np.zeros((n_trees, va_id.size(), max_leaves, max_leaves, 2**scaling))
    lm_si = np.zeros((n_trees, va_id.size(), max_leaves, max_leaves, 2**scaling))

    for b in range(n_trees):
        nb_leaf = leaves_nb[b]
        for leaf_numb in prange(nb_leaf, nogil=True, schedule='dynamic'):
            node_id_v2[leaf_numb].clear()

            if C[0].size() != 0:
                C_b[leaf_numb] = C
                for nv in range(node_idx_trees[b][leaf_numb].size()):
                    add = 0
                    for ns in range(remove_va.size()):
                        if node_idx_trees[b][leaf_numb][nv] == remove_va[ns]:
                            add = 1
                            buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][nv]
                            node_id_v2[leaf_numb].push_back(buff[leaf_numb])
                            break
                    if add == 0:
                        for ci in range(C_b[leaf_numb].size()):
                            for cj in range(C_b[leaf_numb][ci].size()):
                                if C_b[leaf_numb][ci][cj] == node_idx_trees[b][leaf_numb][nv]:
                                    add = 1
                                    node_id_v2[leaf_numb].push_back(C_b[leaf_numb][ci])
                                    break
                            if add == 1:
                                std_remove[vector[vector[long]].iterator, vector[long]](C_b[leaf_numb].begin(), C_b[leaf_numb].end(), C_b[leaf_numb][ci])
                                C_b[leaf_numb].pop_back()
                                break
                node_id[leaf_numb] = node_id_v2[leaf_numb]
            else:
                node_id[leaf_numb].clear()
                for i in range(node_idx_trees[b][leaf_numb].size()):
                    buff[leaf_numb][0] = node_idx_trees[b][leaf_numb][i]
                    node_id[leaf_numb].push_back(buff[leaf_numb])


            for va in range(va_id.size()):
                node_id_va[va][leaf_numb] = node_id[leaf_numb]
                if not std_find[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va]) != node_id_va[va][leaf_numb].end():
                    continue

                std_remove[vector[vector[long]].iterator, vector[long]](node_id_va[va][leaf_numb].begin(), node_id_va[va][leaf_numb].end(), va_id[va])
                node_id_va[va][leaf_numb].pop_back()

                set_size = node_id_va[va][leaf_numb].size()
                pow_set_size = 2**set_size

                for counter in range(0, pow_set_size):
                    va_size = 0
                    S_size = 0
                    for ci in range(set_size):
                        if((counter & (1 << ci)) > 0):
                            for cj in range(node_id_va[va][leaf_numb][ci].size()):
                                S[va, counter, leaf_numb, S_size] = node_id_va[va][leaf_numb][ci][cj]
                                S_size = S_size + 1
                            va_size = va_size + 1

                    for leaf_n in range(nb_leaf):
                        for i in range(data.shape[0]):
                            a_it = 0
                            for s in range(data.shape[1]):
                                if (data[i, s] <= partition_leaves_trees[b, leaf_n, s, 1]) and (data[i, s] > partition_leaves_trees[b, leaf_n, s, 0]):
                                    a_it = a_it + 1
                            if a_it == data.shape[1]:
                                lm_n[b, va, leaf_numb, leaf_n, counter] = lm_n[b, va, leaf_numb, leaf_n, counter] + 1

                            b_it = 0
                            for s in range(S_size):
                                if ((data[i, S[va, counter, leaf_numb, s]] <= partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 1]) * (data[i, S[va, counter, leaf_numb, s]] > partition_leaves_trees[b, leaf_n, S[va, counter, leaf_numb, s], 0])):
                                    b_it = b_it + 1

                            if b_it == S_size:
                                lm_s[b, va, leaf_numb, leaf_n, counter] = lm_s[b, va, leaf_numb, leaf_n, counter] + 1

                                nv_bool = 0
                                for nv in range(va_id[va].size()):
                                    if ((data[i, va_id[va][nv]] > partition_leaves_trees[b, leaf_n, va_id[va][nv], 1]) or (data[i, va_id[va][nv]] <= partition_leaves_trees[b, leaf_n, va_id[va][nv], 0])):
                                        nv_bool = nv_bool + 1
                                        continue

                                if nv_bool == 0:
                                    lm_si[b, va, leaf_numb, leaf_n, counter] = lm_si[b, va, leaf_numb, leaf_n, counter] + 1


    return lm_n, lm_s, lm_si




from acv_app.colors import _colors as colors
import pandas as pd
import numpy as np
import streamlit as st
import matplotlib.pyplot as plt
from acv_explainers.utils import extend_partition

labels = {
    'MAIN_EFFECT': "SHAP main effect value for\n%s",
    'INTERACTION_VALUE': "SHAP interaction value",
    'INTERACTION_EFFECT': "SHAP interaction value for\n%s and %s",
    'VALUE': "SHAP value (impact on model output)",
    'GLOBAL_VALUE': "mean(|SHAP value|) (average impact on model output magnitude)",
    'VALUE_FOR': "SHAP value for\n%s",
    'PLOT_FOR': "SHAP plot for %s",
    'FEATURE': "Feature %s",
    'FEATURE_VALUE': "Feature value",
    'FEATURE_VALUE_LOW': "Low",
    'FEATURE_VALUE_HIGH': "High",
    'JOINT_VALUE': "Joint SHAP value",
    'MODEL_OUTPUT': "Model output value"
}


def write_pg(x_train, x_test, y_train, y_test, acvtree):
    st.sidebar.title("Parameters")
    if y_test.dtype == int or y_test.dtype == bool:
        CLASSIFIER = st.sidebar.checkbox('Classifier', value=True)
    else:
        CLASSIFIER = st.sidebar.checkbox('Classifier', value=False)

    col1, col2, col3 = st.columns(3)

    with col1:
        nb = st.sidebar.number_input(label='SAMPLE SIZE', value=10, min_value=5, max_value=500)
    with col2:
        pi_level = st.sidebar.number_input(label='SDP MIN (\pi)', value=0.9, min_value=0.7, max_value=1.)
    with col3:
        t = st.sidebar.number_input(label='SDP THRESHOLD FOR REGRESSOR', value=10., min_value=1., max_value=500.)

    idx = st.selectbox(
        'Choose the observation you want to explain',
        list(range(nb))
    )

    @st.cache(allow_output_mutation=True)
    def compute_sdp(nb, x_train, y_train, x_test, y_test, pi_level, t):
        sufficient_coal, sdp_coal, sdp_global = acvtree.sufficient_expl_rf(x_test[:nb], y_test[:nb], x_train, y_train,
                                                                           stop=False, pi_level=pi_level,
                                                                           t=t)
        for i in range(len(sufficient_coal)):
            sufficient_coal[i].pop(0)
            sdp_coal[i].pop(0)
        return sufficient_coal, sdp_coal, sdp_global

    @st.cache(allow_output_mutation=True)
    def compute_sdp_rule(obs, x_train_np, y_train_np, x_test_np, y_test_np, t, S):
        sdp, rules = acvtree.compute_sdp_rule(x_test_np[obs:obs+1], y_test_np[obs:obs+1],
                                              x_train_np, y_train_np, S=[S], t=t)
        rule = rules[0]
        columns = [x_train.columns[i] for i in range(x_train.shape[1])]
        rule_string = ['{} <= {} <= {}'.format(rule[i, 0] if rule[i, 0] > -1e+10 else -np.inf, columns[i],
                                               rule[i, 1] if rule[i, 1] < 1e+10 else +np.inf) for i in S]
        rule_string = ' and '.join(rule_string)
        return rule_string

    @st.cache(allow_output_mutation=True)
    def compute_sdp_maxrule(obs, x_train_np, y_train_np, x_test_np, y_test_np, t, S, pi):
        sdp, rules, sdp_all, rules_data, w = acvtree.compute_sdp_maxrules(x_test_np[obs:obs + 1], y_test_np[obs:obs + 1],
                                              x_train_np, y_train_np, S=[S], t=t, pi_level=pi)

        acvtree.fit_global_rules(x_train_np, y_train_np, rules, [S])

        # extend_partition(rules, rules_data, sdp_all, pi=pi, S=[S])

        rule = rules[0]
        columns = [x_train.columns[i] for i in range(x_train.shape[1])]
        rule_string = ['{} <= {} <= {}'.format(rule[i, 0] if rule[i, 0] > -1e+10 else -np.inf, columns[i],
                                               rule[i, 1] if rule[i, 1] < 1e+10 else +np.inf) for i in S]
        rule_string = ' and '.join(rule_string)
        return rule_string

    @st.cache(allow_output_mutation=True)
    def transform_scoal_to_col(sufficient_coal, columns_names):
        col_byobs = []
        for obs in sufficient_coal:
            col = []
            for S in obs:
                name = ''
                for i in range(len(S)):
                    if i != len(S) - 1:
                        name += columns_names[S[i]] + ' - '
                    else:
                        name += columns_names[S[i]]
                col.append(name)
            col_byobs.append(col)
        return col_byobs

    @st.cache(allow_output_mutation=True)
    def compute_local_sdp(idx, sufficient_coal):
        flat = [item for sublist in sufficient_coal[idx] for item in sublist]
        flat = pd.Series(flat)
        flat = dict(flat.value_counts() / len(sufficient_coal[idx]))
        local_sdp = np.zeros(x_train.shape[1])
        for key in flat.keys():
            local_sdp[key] = flat[key]
        return local_sdp

    @st.cache(allow_output_mutation=True)
    def color_max(data, sdp_index):
        color = []
        for i in range(x_train.shape[1]+1):
            if i in sdp_index:
                color.append('background-color: #3e82fc')
            else:
                color.append('')
        color.append('background-color: #ff073a')
        return color

    @st.cache(allow_output_mutation=True)
    def bar_legacy(shap_values, features=None, feature_names=None, max_display=None, show=True):
        # unwrap pandas series

        fig = plt.figure()

        if str(type(features)) == "<class 'pandas.core.series.Series'>":
            if feature_names is None:
                feature_names = list(features.index)
            features = features.values

        if feature_names is None:
            feature_names = np.array([labels['FEATURE'] % str(i) for i in range(len(shap_values))])

        if max_display is None:
            max_display = 7
        else:
            max_display = min(len(feature_names), max_display)

        feature_order = np.argsort(-np.abs(shap_values))

        #
        feature_inds = feature_order[:max_display]
        y_pos = np.arange(len(feature_inds), 0, -1)
        plt.barh(
            y_pos, shap_values[feature_inds],
            0.7, align='center',
            color=[colors.red_rgb if shap_values[feature_inds[i]] < 0 else colors.blue_rgb for i in range(len(y_pos))]
        )
        for y in range(len(y_pos)):
            plt.text(shap_values[feature_inds][y] + 0.001, y_pos[y] - 0.07, round(shap_values[feature_inds][y], 3))

        plt.yticks(y_pos, fontsize=13)
        if features is not None:
            features = list(features)

            # try and round off any trailing zeros after the decimal point in the feature values
            for i in range(len(features)):
                try:
                    if round(features[i]) == features[i]:
                        features[i] = int(features[i])
                except TypeError:
                    pass  # features[i] must not be a number
        yticklabels = []
        for i in feature_inds:
            if features is not None:
                yticklabels.append(feature_names[i] + " = " + str(features[i]))
            else:
                yticklabels.append(feature_names[i])
        plt.gca().set_yticklabels(yticklabels)
        plt.gca().xaxis.set_ticks_position('bottom')
        plt.gca().yaxis.set_ticks_position('none')
        plt.gca().spines['right'].set_visible(False)
        plt.gca().spines['top'].set_visible(False)
        # plt.gca().spines['left'].set_visible(False)

        plt.xlabel("Frequency of apparition in the Sufficient Coalitions")
        return fig

    # explantions_load_state = st.text('Computing SDP explanations...')
    sufficient_coal, sdp_coal, sdp_global = compute_sdp(nb, x_train.values.astype(np.double),
                                                        y_train.astype(np.double), x_test.values.astype(np.double),
                                                        y_test.astype(np.double), pi_level=pi_level, t=t)

    sufficient_coal_names = transform_scoal_to_col(sufficient_coal, x_train.columns)
    # explantions_load_state.text("SDP explanation Done!")

    # st.subheader('All sufficient coalitions')

    if len(sufficient_coal[idx]) == 0:
        st.text('No explanation was found for this observation')
    else:

        col1, col2 = st.columns(2)
        with col1:
            st.header('All sufficient explanations')
            sufficient_coal_df = {'Sufficient explanations': sufficient_coal_names[idx],
                                  'SDP': sdp_coal[idx]}

            sufficient_coal_df = pd.DataFrame(sufficient_coal_df)
            # print(sufficient_coal_df.head())
            st.dataframe(sufficient_coal_df, 6000, 6000)

        with col2:
            st.header('Local Explanatory Importance')
            local_sdp = compute_local_sdp(idx, sufficient_coal)
            # data = {'feature_names': [x_train.columns[i] for i in range(x_train.shape[1])],
            #         'feature_importance': local_sdp}
            # fi_df = pd.DataFrame(data)

            # Sort the DataFrame in order decreasing feature importance
            # fi_df.sort_values(by=['feature_importance'], ascending=False, inplace=True)
            # sns.set_theme(font='sans-serif')
            fig = bar_legacy(local_sdp, x_test.values[idx], x_test.columns)
            # sns.set(font_scale=1.5)
            # sns.set_theme(font='sans-serif')
            # sns.barplot(x=fi_df['feature_importance'], y=fi_df['feature_names'], color='#3e82fc')
            # plt.xlabel('Frequency of apparition in the Sufficient Coalitions')
            # for y in range(len(fi_df['feature_importance'].values)):
            #     plt.text(fi_df['feature_importance'].values[y], y, round(fi_df['feature_importance'].values[y], 3))
            #
            # plt.ylabel(' ')
            st.pyplot(fig)

        st.header('Feature values highlight by SDP')
        st.text('This observation has {} different explanations, below to observe their values'.format(
            len(sufficient_coal[idx])))

        exp_idx = st.selectbox(
            'Change the explanations',
            list(range(len(sufficient_coal[idx])))
        )

        x_group = pd.DataFrame(x_test.values[idx:idx + 1], columns=x_test.columns)
        x_group['Output'] = y_test[idx]
        x_group['Same Decision Probability (SDP)'] = sdp_coal[idx][exp_idx]
        st.dataframe(x_group.iloc[:1].style.apply(color_max, sdp_index=sufficient_coal[idx][exp_idx], axis=1))

        st.header('Local rule explanation')
        rule_string = compute_sdp_rule(idx, x_train.values.astype(np.double), y_train.astype(np.double),
                                       x_test.values.astype(np.double), y_test.astype(np.double), t, sufficient_coal[idx][exp_idx])
        # st.markdown(rule_string)

        st.markdown("<" + 'h3' + " style='text-align: " + \
                    "; color:" + 'black' + "; '>" + rule_string + "</" + 'h3' + ">",
                    unsafe_allow_html=True)

        st.header('Sufficient local rule explanation (Maximal rule)')
        maxrule = st.checkbox('Compute', value=False)

        if maxrule:
            rule_string = compute_sdp_maxrule(idx, x_train.values.astype(np.double), y_train.astype(np.double),
                                           x_test.values.astype(np.double), y_test.astype(np.double), t,
                                           sufficient_coal[idx][exp_idx], pi_level)

            st.markdown("<" + 'h3' + " style='text-align: " + \
                        "; color:" + 'black' + "; '>" + rule_string + "</" + 'h3' + ">",
                        unsafe_allow_html=True)
            # st.markdown(rule_string)

            rule_info = {'Rule coverage': acvtree.rules_coverage,
                          'Rule accuracy/mse': acvtree.rules_acc}

            rule_info = pd.DataFrame(rule_info)
            st.dataframe(rule_info, 6000, 6000)



from .utils import *
import numpy as np
from tqdm import tqdm
from scipy.special import comb


def shap_values_leaves(x, partition_leaves, data, node_idx, leaf_idx, leaves_nb, weight_samples, v, C,
                       num_outputs):
    """
    Compute SV using multi-game algorithm
    Args:
        x (np.array): Input of shape = (# samples, # features)
        partition_leaves (np.array): It contains the hype-rectangle of each leaves, shape = # leaves x D x 2
        data_leaves (np.array): Boolean matrix of shape = # leaves x D * N, if data_leaves[i, j, z ] = 1 means that obs
        z fall in rectangle j of leaf i.
        node_idx (list): node_idx[i] is the indexes of the nodes that it is in the path of leaf i
        leaf_idx (list): Index of the leaves nodes
        weight_samples (array): weight_samples[i] is the number of samples that fall in leaf node i
        v (array): v[i] is the value of node i, shape = # nodes x num_outputs
        C (list): Indexes of the variables group together (coalition variables)
        num_outputs (int): size of the output

    Returns:
            (array): SV with multi-game algorithm, Shape = N x D x num_outputs
    """
    phi = np.zeros((x.shape[0], x.shape[1], num_outputs)).astype(np.float64)

    for leaf_numb in range(leaves_nb):
        leaf_id = leaf_idx[leaf_numb]
        partition_leaf = partition_leaves[leaf_numb]
        node_id = node_idx[leaf_numb]
        d = x.shape[1]
        va_id = list(range(d))

        lm = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                             for s in range(d)], axis=0))
        # start handle coalition
        if C[0] != []:
            for c in C:
                d -= len(c)
                va_id = list(set(va_id) - set(c))
            for c in C:
                va_id += [c]
                present = np.sum([c_i in node_id for c_i in c])
                if present != 0:
                    node_id = node_id + c
            d += len(C)

        for i in va_id:
            if not set(convert_list(i)).issubset(node_id):
                continue

            Sm = list(set(node_id) - set(convert_list(i)))

            if C[0] != []:
                buffer_Sm = Sm.copy()
                for c in C:
                    if set(c).issubset(buffer_Sm):
                        Sm = list(set(Sm) - set(c))
                for c in C:
                    if set(c).issubset(buffer_Sm):
                        Sm += [c]
            # end handle coalition

            for S in powerset(Sm):

                csi = np.prod([(x[:, s] <= partition_leaf[s, 1]) * (x[:, s] > partition_leaf[s, 0])
                               for s in chain_l(S) + convert_list(i)], axis=0)
                cs = np.prod([(x[:, s] <= partition_leaf[s, 1]) * (x[:, s] > partition_leaf[s, 0])
                              for s in chain_l(S)], axis=0)

                coef = 0
                for l in range(1, d - len(Sm)):
                    coef += comb(d - len(Sm) - 1, l, exact=True) * comb(d - 1, l + len(S), exact=True) ** (-1)

                lm_si = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                                        for s in chain_l(S) + convert_list(i)], axis=0))

                lm_s = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                                       for s in chain_l(S)], axis=0))

                p_si = (csi * lm) / lm_si

                if len(S) == 0:
                    p_s = lm / data.shape[0]
                else:
                    p_s = (cs * lm) / lm_s
                for j in convert_list(i):
                    phi[:, j, :] += (comb(d - 1, len(S), exact=True) ** (-1) + coef) * (p_si - p_s)[:, None] * v[
                                                                                                                   leaf_id][
                                                                                                               None, :]
    return phi / d


def shap_values_leaves_v2(x, partition_leaves, data, node_idx, leaf_idx, leaves_nb, weight_samples, v, C,
                          num_outputs):
    """
    Compute SV using multi-game algorithm
    Args:
        x (np.array): Input of shape = (# samples, # features)
        partition_leaves (np.array): It contains the hype-rectangle of each leaves, shape = # leaves x D x 2
        data_leaves (np.array): Boolean matrix of shape = # leaves x D * N, if data_leaves[i, j, z ] = 1 means that obs
        z fall in rectangle j of leaf i.
        node_idx (list): node_idx[i] is the indexes of the nodes that it is in the path of leaf i
        leaf_idx (list): Index of the leaves nodes
        weight_samples (array): weight_samples[i] is the number of samples that fall in leaf node i
        v (array): v[i] is the value of node i, shape = # nodes x num_outputs
        C (list): Indexes of the variables group together (coalition variables)
        num_outputs (int): size of the output

    Returns:
            (array): SV with multi-game algorithm, Shape = N x D x num_outputs
    """
    phi = np.zeros((x.shape[0], x.shape[1], num_outputs)).astype(np.float64)

    if C[0] != []:
        coal_va = [C[ci][cj] for ci in range(len(C)) for cj in range(len(C[ci]))]
        remove_va = [[i] for i in range(x.shape[1]) if i not in coal_va]
        va_id = remove_va + C
    else:
        va_id = [[i] for i in range(x.shape[1])]
        remove_va = [[i] for i in range(x.shape[1])]

    d = len(va_id)

    for leaf_numb in range(leaves_nb):
        leaf_id = leaf_idx[leaf_numb]
        partition_leaf = partition_leaves[leaf_numb]
        node_id = node_idx[leaf_numb]

        lm = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                             for s in range(data.shape[1])], axis=0))
        # start handle coalition
        node_id_v2 = []
        C_b = C.copy()
        if C[0] != []:
            for va in node_id:
                if va in remove_va:
                    node_id_v2 += [[va]]
                else:
                    for ci in C_b:
                        if va in ci:
                            node_id_v2 += [ci]
                            C_b.remove(ci)
                            break
        else:
            node_id_v2 = [[i] for i in node_id]

        node_id = node_id_v2.copy()
        # end handle coalition
        for i in va_id:
            if not i in node_id:
                continue

            Sm = node_id.copy()
            Sm.remove(i)
            for S in powerset(Sm):
                S_size = len(S)
                S = sum(list(S), [])
                Si = S + i

                csi = np.prod([(x[:, s] <= partition_leaf[s, 1]) * (x[:, s] > partition_leaf[s, 0])
                               for s in Si], axis=0)
                cs = np.prod([(x[:, s] <= partition_leaf[s, 1]) * (x[:, s] > partition_leaf[s, 0])
                              for s in S], axis=0)

                coef = 0
                for l in range(1, d - len(Sm)):
                    coef += comb(d - len(Sm) - 1, l, exact=True) * comb(d - 1, l + S_size, exact=True) ** (-1)

                lm_si = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                                        for s in Si], axis=0))

                lm_s = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                                       for s in S], axis=0))

                p_si = (csi * lm) / lm_si

                if len(S) == 0:
                    p_s = lm / data.shape[0]
                else:
                    p_s = (cs * lm) / lm_s
                for j in convert_list(i):
                    phi[:, j, :] += (comb(d - 1, S_size, exact=True) ** (-1) + coef) * (p_si - p_s)[:, None] * v[
                                                                                                                   leaf_id][
                                                                                                             None, :]
    return phi / d


def shap_values_acv_leaves(x, partition_leaves, data, node_idx, leaf_idx, leaves_nb, weight_samples, v, C,
                           S_star, N_star,
                           num_outputs):
    """
        Compute ACV-SV using multi-game algorithm
        Args:
            x (np.array): Input of shape = (# samples, # features)
            partition_leaves (np.array): It contains the hype-rectangle of each leaves, shape = # leaves x D x 2
            data_leaves (np.array): Boolean matrix of shape = # leaves x D * N, if data_leaves[i, j, z ] = 1 means that obs
            z fall in rectangle j of leaf i.
            node_idx (list): node_idx[i] is the indexes of the nodes that it is in the path of leaf i
            leaf_idx (list): Index of the leaves nodes
            weight_samples (array): weight_samples[i] is the number of samples that fall in leaf node i
            v (array): v[i] is the value of node i, shape = # nodes x num_outputs
            C (list): Indexes of the variables group together (coalition variables)
            num_outputs (int): size of the output
            S_star (list): Indexes of the Active variables, see section ACV of the papers
            N_star (list): Indexes of the Null variables, see section ACV of papers

        Returns:
                (array): ACV-SV with multi-game algorithm, Shape = N x D x num_outputs
        """
    phi = np.zeros((x.shape[0], x.shape[1], num_outputs)).astype(np.float64)
    for leaf_numb in range(leaves_nb):
        leaf_id = leaf_idx[leaf_numb]
        partition_leaf = partition_leaves[leaf_numb]
        # data_leaf = data_leaves[leaf_numb]
        node_id = node_idx[leaf_numb]
        lm = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                             for s in range(data.shape[1])], axis=0))

        va_id = S_star
        d = len(va_id)
        node_id = list(set(node_id) - set(N_star))
        # start handle coalition
        if C[0] != []:
            for c in C:
                if set(c).issubset(S_star):
                    d -= len(c)
                    va_id = list(set(va_id) - set(c))
            for c in C:
                if set(c).issubset(S_star):
                    va_id += [c]
                    present = np.sum([c_i in node_id for c_i in c])
                    if present != 0:
                        node_id = node_id + c
                    d += 1

        for i in va_id:

            if not set(convert_list(i)).issubset(node_id):
                comp_si = np.prod([(x[:, s] <= partition_leaf[s, 1]) * (x[:, s] >= partition_leaf[s, 0])
                                   for s in N_star], axis=0)

                lm_si = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                             for s in N_star], axis=0))

                p_si = comp_si * lm / lm_si
                p_s = lm / data.shape[0]
                for a in convert_list(i):
                    phi[:, a] += ((p_si - p_s)[:, None] * v[leaf_id][None, :])
                continue


            Sm = list(set(node_id) - set(convert_list(i)))
            buffer_Sm = Sm.copy()

            if C[0] != []:
                for c in C:
                    if set(c).issubset(S_star) and set(c).issubset(buffer_Sm):
                        Sm = list(set(Sm) - set(c))
                for c in C:
                    if set(c).issubset(S_star) and set(c).issubset(buffer_Sm):
                        Sm += [c]
            for S in powerset(Sm):

                comp_si = np.prod([(x[:, s] <= partition_leaf[s, 1]) * (x[:, s] >= partition_leaf[s, 0])
                                   for s in chain_l(S) + convert_list(i) + N_star], axis=0)
                comp_s = np.prod([(x[:, s] <= partition_leaf[s, 1]) * (x[:, s] >= partition_leaf[s, 0])
                                  for s in chain_l(S) + N_star], axis=0)

                coef = 0
                for l in range(1, d - len(Sm)):
                    coef += comb(d - len(Sm) - 1, l) * comb(d - 1, l + len(S)) ** (-1)

                lm_si = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                             for s in chain_l(S) + convert_list(i) + N_star], axis=0))
                lm_s = np.sum(np.prod([(data[:, s] <= partition_leaf[s, 1]) * (data[:, s] > partition_leaf[s, 0])
                             for s in chain_l(S) + N_star], axis=0))

                p_si = comp_si * lm / lm_si
                p_s = comp_s * lm / lm_s
                if len(S) != 0:
                    for a in convert_list(i):
                        phi[:, a] += (
                                (comb(d - 1, len(S)) ** (-1) + coef) * (p_si - p_s)[:, None] * v[leaf_id][None, :])
                else:
                    P_soff = lm / data.shape[0]
                    for a in convert_list(i):
                        phi[:, a] += ((p_si - P_soff)[:, None] * v[leaf_id][None, :] + coef * (p_si - p_s)[:, None] * v[
                                                                                                                          leaf_id][
                                                                                                                      None,
                                                                                                                      :])

    return phi / d


def cond_sdp_forest_clf(x, fx, tx, forest, S, data):
    """
    Compute SDP(x, forest_classifier) of variables in S
    Args:
        x (array): observation
        fx (float): forest(x)
        tx (float): threshold of the classifier
        forest (All TreeBased models): model
        S (list): index of variables on which we want to compute the SDP
        data (array): data used to compute the SDP

    Returns:
        float: SDP(x, forest_classifier)
    """
    if len(S) == len(x):
        return 1
    elif S == []:
        return 0

    n_trees = len(forest)
    names = ['all', 'up', 'down']
    part_forest = [get_tree_partition(x, fx, tx, forest[i], S, data=data, is_reg=False) for i in
                   range(n_trees)]
    sdp = 0
    for i in range(n_trees):
        mean_forest = {'all': 0, 'up': 0, 'down': 0}
        value = part_forest[i][1]
        for name in names:
            p = part_forest[i][0][name].sum(axis=1).astype(np.float64)
            p_s = part_forest[i][0]['s_{}'.format(name)].sum(axis=1).astype(np.float64)
            prob = np.true_divide(p, p_s, out=np.zeros_like(p), where=p_s != 0)
            # if np.sum(prob) != 0:
            #     prob = prob / np.sum(prob)

            mean_forest[name] += np.sum(prob[:, None] * value, axis=0)

        s = (mean_forest['all'] - mean_forest['down']) / (mean_forest['up'] - mean_forest['down'])
        sdp += 0 * (s[int(fx)] < 0) + 1 * (s[int(fx)] > 1) + s[int(fx)] * (0 <= s[int(fx)] <= 1)
    # sdp = 0 * (sdp[int(fx)] < 0) + 1 * (sdp[int(fx)] > 1) + sdp[int(fx)] * (0 <= sdp[int(fx)] <= 1)
    return sdp / n_trees


def cond_sdp_forest(x, fx, tx, forest, S, data):
    """
        Compute SDP_S(x, treeBased models) of variables in S

        Args:
            x (array): observation
            fx (float): forest(x)
            tx (float): threshold of the classifier
            forest (All TreeBased models): model
            S (list): index of variables on which we want to compute the SDP
            data (array): data used to compute the SDP

        Returns:
            float: SDP_S(x, treeBased models)
        """
    n_trees = len(forest)
    names = ['all', 'up', 'down']
    part_forest = [get_tree_partition(x, fx, tx, forest[i], S, data=data) for i in
                   range(n_trees)]
    mean_forest = {'all': 0, 'up': 0, 'down': 0}
    for i in range(n_trees):
        for j in range(n_trees):
            if i == j:
                value = part_forest[i][1]
                for name in names:
                    p = part_forest[i][0][name].sum(axis=1).astype(np.float64)
                    p_s = part_forest[i][0]['s_{}'.format(name)].sum(axis=1).astype(np.float64)
                    prob = np.true_divide(p, p_s, out=np.zeros_like(p), where=p_s != 0)
                    # if np.sum(prob) != 0:
                    #     prob = prob / np.sum(prob)

                    mean_forest[name] += (np.sum(prob * value ** 2)) / (n_trees ** 2) - (
                            2 * fx * np.sum(prob * value)) / n_trees

            else:

                value_i = part_forest[i][1].reshape(-1, 1)
                value_j = part_forest[j][1].reshape(-1, 1)
                value_ij = np.matmul(value_i, value_j.T)

                for name in names:
                    p_i = part_forest[i][0][name]
                    p_j = part_forest[j][0][name]
                    p_ij = np.matmul(p_i, p_j.T).astype(np.float64)
                    s_ij = np.matmul(part_forest[i][0]['s_{}'.format(name)],
                                     part_forest[j][0]['s_{}'.format(name)].T).astype(np.float64)

                    prob_ij = np.true_divide(p_ij, s_ij, out=np.zeros_like(p_ij), where=s_ij != 0)
                    # if np.sum(prob_ij) != 0:
                    #     prob_ij = prob_ij / np.sum(prob_ij)

                    mean_forest[name] += np.sum(np.multiply(prob_ij, value_ij)) / n_trees ** 2

    sdp = (mean_forest['up'] - mean_forest['all']) / (mean_forest['up'] - mean_forest['down'])
    sdp = 0 * (sdp < 0) + 1 * (sdp > 1) + sdp * (0 <= sdp <= 1)
    return sdp


def swing_tree_shap(X, tX, threshold, data, C, value_function):
    N = X.shape[0]
    m = X.shape[1]
    va_id = list(range(m))
    va_buffer = va_id.copy()
    if C[0] != []:
        for c in C:
            m -= len(c)
            va_id = list(set(va_id) - set(c))
        m += len(C)
        for c in C:
            va_id += [c]

    phi = np.zeros(shape=(X.shape[0], X.shape[1]))

    swings = np.zeros((N, X.shape[1], 2))
    swings_prop = np.zeros((N, X.shape[1], 3))

    for i in va_id:
        Sm = list(set(va_buffer) - set(convert_list(i)))

        if C[0] != []:
            buffer_Sm = Sm.copy()
            for c in C:
                if set(c).issubset(buffer_Sm):
                    Sm = list(set(Sm) - set(c))
            for c in C:
                if set(c).issubset(buffer_Sm):
                    Sm += [c]

        for S in powerset(Sm):
            weight = comb(m - 1, len(S)) ** (-1)
            v_plus = value_function(X=X, tX=tX, threshold=threshold, data=data, S=chain_l(S) + convert_list(i))
            v_minus = value_function(X=X, tX=tX, threshold=threshold, data=data, S=chain_l(S))

            dif_pos = (v_plus - v_minus) > 0
            dif_neg = (v_plus - v_minus) < 0
            dif_null = (v_plus - v_minus) == 0
            value = ((v_plus - v_minus) * weight) / m

            for a in convert_list(i):
                phi[:, a] += weight * (v_plus - v_minus)

                swings[:, a, 0] += dif_pos * value
                swings[:, a, 1] += dif_neg * value

                swings_prop[:, a, 0] += dif_pos
                swings_prop[:, a, 1] += dif_neg
                swings_prop[:, a, 2] += dif_null

    return phi / m, swings, swings_prop


def swing_tree_shap_clf(X, data, C, value_function, threshold=0.9):
    N = X.shape[0]
    m = X.shape[1]
    va_id = list(range(m))
    va_buffer = va_id.copy()
    if C[0] != []:
        for c in C:
            m -= len(c)
            va_id = list(set(va_id) - set(c))
        m += len(C)
        for c in C:
            va_id += [c]

    phi = np.zeros(shape=(X.shape[0], X.shape[1]))

    swings = np.zeros((N, X.shape[1], 2))
    swings_prop = np.zeros((N, X.shape[1], 3))

    for i in va_id:
        Sm = list(set(va_buffer) - set(convert_list(i)))

        if C[0] != []:
            buffer_Sm = Sm.copy()
            for c in C:
                if set(c).issubset(buffer_Sm):
                    Sm = list(set(Sm) - set(c))
            for c in C:
                if set(c).issubset(buffer_Sm):
                    Sm += [c]

        for S in tqdm(powerset(Sm)):
            weight = comb(m - 1, len(S)) ** (-1)
            v_plus = 1.*(value_function(X=X, data=data, S=np.array(chain_l(S) + convert_list(i)).astype(np.long)) >= threshold)
            v_minus = 1.*(value_function(X=X, data=data, S=np.array(chain_l(S)).astype(np.long)) >= threshold)

            dif_pos = (v_plus - v_minus) > 0
            dif_neg = (v_plus - v_minus) < 0
            dif_null = (v_plus - v_minus) == 0
            value = ((v_plus - v_minus) * weight) / m

            for a in convert_list(i):
                phi[:, a] += weight * (v_plus - v_minus)

                swings[:, a, 0] += dif_pos * value
                swings[:, a, 1] += dif_neg * value

                swings_prop[:, a, 0] += dif_pos
                swings_prop[:, a, 1] += dif_neg
                swings_prop[:, a, 2] += dif_null

    return phi / m, swings, swings_prop


def local_sdp(x, threshold, proba, index, data, final_coal, decay, C, verbose, cond_func):
    """
    Find the Sufficient coalition S* at level "proba", then recompute recursively S* by decreasing the "proba" with
    "decay" for all tree-based models

    Args:
        x (np.array): observation
        f (float): forest(x)
        forest (All TreeBased models): model
        threshold (float): threshold of the classification in (0, 1)
        proba (float): the level of the Sufficient Coalition \pi
        index (list): index of the variables of x
        data (array): data used for the estimation
        final_coal (list): the list that will contain the cluster find by SDP, empty [] at initialization
        decay (float): the probability decay used in the recursion step
        C (list[list]): list of the coalition of variable by their index, default value when no coalition is [[]]

    Returns:
        (list[list]): list of the cluster find, the first is the Sufficient Coalition S* of level "proba". The remaining
        is the Sufficient Coalition when we remove previous S*, set proba = decay * proba and recursively apply the
        function.
    """
    C_off = [()]
    va_id = index.copy()
    m = len(index)

    if C[0] != []:
        for c in C:
            if set(c).issubset(va_id):
                va_id = list(set(va_id) - set(c))
                m -= len(c)
        for c in C:
            if set(c).issubset(index):
                va_id += [c]
                m += 1

    c_value = {i: {} for i in range(m + 1)}
    i_best, c_best = {i: -np.inf for i in range(m + 1)}, {i: -np.inf for i in range(m + 1)}
    find = False

    if len(va_id) == 0:
        if verbose > 0:
            print('End')
        return final_coal
    else:
        for size in range(len(va_id) + 1):
            for c in itertools.combinations(va_id, size):
                if c not in C_off:

                    value = cond_func(x, threshold, S=chain_l(c), data=data)[0]
                    c_value[size][str(c)] = value

                    # if c_value[size][str(c)] < proba:
                    #     C_off.append(c)

                    if c_value[size][str(c)] >= proba and c_value[size][str(c)] >= i_best[size]:
                        i_best[size] = c_value[size][str(c)]
                        c_best[size] = chain_l(c)

            if c_best[size] != -np.inf:
                find = True
                final_coal.append(c_best[size])
                threshold = threshold * decay
                break

            elif len(chain_l(c)) == len(va_id):
                find = True
                final_coal.append(chain_l(tuple(va_id)))

        if find == True:
            for c_i in final_coal[-1]:
                index.remove(c_i)
            if verbose > 0:
                print('REMAINING VARIABLE = {}'.format(index))

            return local_sdp(x, threshold, proba, index, data, final_coal, decay, C, verbose, cond_func)


def global_sdp_importance(data, data_bground, columns_names, pi_level, decay, threshold,
                          proba, C, verbose, cond_func):
    """
    Compute the Global SDP across "data" at level "pi_level" for forest_regressor.

    Args:
        forest (RandomForestRegressor): model
        data (array): data used to compute the Global SDP
        data_bground (array): data used in the estimations of the SDP
        columns_names (list): names of the variables
        pi_level (float): proba used for the selection criterion. We count each time for a variable if it is on a
                              set with SDP >= global proba
        decay (float): decay value used when recursively apply the local_sdp function .
        threshold (float): the radius t of the SDP regressor (see paper: SDP for regression)
        proba (float): the  level of the Sufficient Coalition
        algo (string): name of the estimator, recommended 'plugin'
        C (list[list]): list of the coalition of variable by their index, default value when no coalition is [[]]

    Returns:
        (array): _
        (array): _
        (array): _
        (array): Global SDP for coalition
        (array): GLobal SDP of each variable
    """
    n_size = len(data)
    sdp_importance = {str(i): [] for i in range(data.shape[1])}
    sdp_coal_proba = {}
    sdp_importance_coal_count = {}
    sdp_importance_variable_count = {str(i): 0 for i in range(data.shape[1])}

    for i in tqdm(range(n_size)):
        final_coal = []
        index = list(range(data.shape[1]))
        ind = np.expand_dims(data[i], 0)

        local_sdp(ind, threshold, proba, index, data_bground, final_coal, decay,
                  C=C, verbose=verbose, cond_func=cond_func)

        for c in final_coal:
            value = cond_func(ind, threshold, S=chain_l(c), data=data)[0]
            #             sdp_imp_name = []

            if len(c) > 1:
                sdp_imp_name = [columns_names[i] for i in c]
            else:
                sdp_imp_name = [columns_names[c[0]]]

            if value >= pi_level:
                if str(sdp_imp_name) in sdp_coal_proba.keys():
                    sdp_coal_proba[str(sdp_imp_name)].append(value)
                    sdp_importance_coal_count[str(sdp_imp_name)] += 1 / n_size
                else:
                    sdp_coal_proba[str(sdp_imp_name)] = [value]
                    sdp_importance_coal_count[str(sdp_imp_name)] = 1 / n_size

                for c_i in c:
                    sdp_importance[str(c_i)].append(value)
                    sdp_importance_variable_count[str(c_i)] += 1 / n_size

    #             for c_i in c:
    #                 sdp_importance[str(c_i)].append(value)
    #                 if value >= global_t:
    #                     sdp_importance_variable_count[str(c_i)] += 1/n_size

    sdp_importance_m = [np.mean(sdp_importance[key]) for key in sdp_importance.keys()]
    sdp_coal_proba = {key: np.mean(sdp_coal_proba[key]) for key in sdp_coal_proba.keys()}

    return sdp_importance_m, sdp_importance, sdp_coal_proba, sdp_importance_coal_count, sdp_importance_variable_count


def shap_values_leaves_notoptimized(X, data, C, tree):
    N = X.shape[0]
    m = X.shape[1]
    va_id = list(range(m))
    va_buffer = va_id.copy()

    if C[0] != []:
        for c in C:
            m -= len(c)
            va_id = list(set(va_id) - set(c))
        m += len(C)
        for c in C:
            va_id += [c]

    phi = np.zeros(shape=(X.shape[0], X.shape[1], tree.values.shape[2]))

    for i in tqdm(va_id):
        Sm = list(set(va_buffer) - set(convert_list(i)))

        if C[0] != []:
            buffer_Sm = Sm.copy()
            for c in C:
                if set(c).issubset(buffer_Sm):
                    Sm = list(set(Sm) - set(c))
            for c in C:
                if set(c).issubset(buffer_Sm):
                    Sm += [c]

        for S in powerset(Sm):
            weight = comb(m - 1, len(S)) ** (-1)
            v_plus = tree.compute_exp_normalized(X=X, S=np.array(chain_l(S) + convert_list(i)).astype(int), data=data)
            v_minus = tree.compute_exp_normalized(X=X, S=np.array(chain_l(S)).astype(int), data=data)

            for a in convert_list(i):
                phi[:, a] += weight * (v_plus - v_minus)

    return phi / m


def shap_values_discrete_notoptimized(X, data, C, tree):
    N = X.shape[0]
    m = X.shape[1]
    va_id = list(range(m))
    va_buffer = va_id.copy()

    if C[0] != []:
        for c in C:
            m -= len(c)
            va_id = list(set(va_id) - set(c))
        m += len(C)
        for c in C:
            va_id += [c]

    phi = np.zeros(shape=(X.shape[0], X.shape[1], tree.values.shape[2]))

    for i in tqdm(va_id):
        Sm = list(set(va_buffer) - set(convert_list(i)))

        if C[0] != []:
            buffer_Sm = Sm.copy()
            for c in C:
                if set(c).issubset(buffer_Sm):
                    Sm = list(set(Sm) - set(c))
            for c in C:
                if set(c).issubset(buffer_Sm):
                    Sm += [c]

        for S in powerset(Sm):
            weight = comb(m - 1, len(S)) ** (-1)
            v_plus = tree.compute_exp_cat(X=X, S=np.array(chain_l(S) + convert_list(i)).astype(int), data=data)
            v_minus = tree.compute_exp_cat(X=X, S=np.array(chain_l(S)).astype(int), data=data)

            for a in convert_list(i):
                phi[:, a] += weight * (v_plus - v_minus)

    return phi / m